import numpy as np
from .Images import rcParams
from .sigma import sigmaBlue, sigmaRed, sigmaGeneral
from .fits import gaussian2DFit
import matplotlib.pyplot as plt
from scipy.constants import *

def numAtomsGeneral(image, delta, wLen, Gamma, s=0, plot=True, p0=None, bounds=[(), ()]):
    '''
    Calculates number of atoms from blue shadow imaging.

    Parameters:
        image: a numpy.ndarray, OD from the experiment
        delta: a float, detuning of the probe in MHz
        wLen: float, wavelength of the probe in nm
        Gamma: flaot, linewidth of the excited state in MHz.
        s (optional): a float, saturation parameter of the probe. Default is 0.
        plot (optional): a bool, flag to plot the gaussian fits if True.
            Default is True.
        p0 (optional): a list, initial guess parameters corresponding to gaussian2D function
        bounds (optional): list of lists, bounds for parameters in the form `[[lower bounds], [upper bounds]]`.
    Returns:
        a tuple, `(number of atoms from 2D gaussian fit,
        number of atoms from pixel sum, number density from gaussian fit,
        sigma_x, sigma_y, amplitude, x0, y0)`
    '''
    scat = sigmaGeneral(delta, wLen, Gamma, s)
    pOpt, pCov = gaussian2DFit(image, p0=p0, bounds=bounds, plot=plot)
    amp, xo, yo, sigma_x, sigma_y, theta, offset = pOpt
    imaging_params = rcParams().params
    pixelSize = imaging_params['pixelSize']
    magnification = imaging_params['magnification']
    binning = imaging_params['binning']
    fac = (pixelSize*binning/magnification)
    NGaussian = 2*pi*amp*sigma_x*sigma_y*fac**2/scat
    NPixel = np.sum(np.sum(image, axis=0), axis=0)*fac**2/scat
    Ndensity = NGaussian/((2*pi*sigma_x*sigma_y*fac**2)**(3/2))
    return NGaussian, NPixel, Ndensity, sigma_x, sigma_y, amp, xo, yo

def numAtomsBlue(image, delta, s=0, plot=True, p0=None, bounds=[(), ()]):
    '''
    Calculates number of atoms from blue shadow imaging.

    Parameters:
        image: a numpy.ndarray, OD from the experiment
        delta: a float, detuning of the probe, 2*(AOMFreq-69) MHz
        s(optional): a float, saturation parameter of the probe. Default is 0.
        plot(optional): a bool, flag to plot the gaussian fits if True.
            Default is True.
        p0 (optional): a list, initial guess parameters corresponding to gaussian2D function
        bounds (optional): list of lists, bounds for parameters in the form `[[lower bounds], [upper bounds]]`.
    Returns:
        a tuple, `(number of atoms from 2D gaussian fit,
        number of atoms from pixel sum, number density from gaussian fit,
        sigma_x, sigma_y, amplitude, x0, y0)`
    '''
    scat = sigmaBlue(delta, 87, s)
    pOpt, pCov = gaussian2DFit(image, p0=p0, bounds=bounds, plot=plot)
    amp, xo, yo, sigma_x, sigma_y, theta, offset = pOpt
    imaging_params = rcParams().params
    pixelSize = imaging_params['pixelSize']
    magnification = imaging_params['magnification']
    binning = imaging_params['binning']
    fac = (pixelSize*binning/magnification)
    NGaussian = 2*pi*amp*sigma_x*sigma_y*fac**2/scat
    NPixel = np.sum(np.sum(image, axis=0), axis=0)*fac**2/scat
    Ndensity = NGaussian/((2*pi*sigma_x*sigma_y*fac**2)**(3/2))
    return NGaussian, NPixel, Ndensity, sigma_x, sigma_y, amp, xo, yo

def numAtomsRed(image, delta, s=0, plot=True,p0=None,bounds=[(), ()]):
    """
    Calculates number of atoms from red shadow imaging.

    Parameters:
        image: a numpy.ndarray, OD from the experiment
        delta: float, detuning of the probe in kHz
        s(optional): a float, saturation parameter of the probe. Default is 0.
        plot(optional): a bool, a flag to plot the gaussian fits if True.
            Default is True.
        p0 (optional): a list, initial guess parameters corresponding to gaussian2D function
        bounds (optional): list of lists, bounds for parameters in the form `[[lower bounds], [upper bounds]]`.
    Returns:
        a tuple, `(number of atoms from 2D gaussian fit,
        number of atoms from pixel sum, number density from gaussian fit,
        sigma_x, sigma_y, amplitude, x0, y0)`
    """
    scat = sigmaRed(delta, s)
    pOpt, pCov = gaussian2DFit(image, p0=p0, bounds=bounds, plot=plot)
    amp, xo, yo, sigma_x, sigma_y, theta, offset = pOpt
    imaging_params = rcParams().params
    pixelSize = imaging_params['pixelSize']
    magnification = imaging_params['magnification']
    binning = imaging_params['binning']
    fac = (pixelSize*binning/magnification)
    NGaussian = 2*pi*amp*sigma_x*sigma_y*fac**2/scat
    NPixel = np.sum(np.sum(image, axis=0), axis=0)*fac**2/scat
    Ndensity = NGaussian/((2*pi*sigma_x*sigma_y*fac**2)**(3/2))
    return NGaussian, NPixel, Ndensity, sigma_x, sigma_y, amp, xo, yo


def temperature(sizes, timeStamps, plot=True):
    '''
    Calculates temperature of the cold gas given size of the cloud at various times of flight.

    Parameters:
        sizes: 2D array, sizes of the cloud extracted from gaussian2D fit in pixel units.
            Eg: ([[x1, y1], [x2, y2], [x3, y3]])
        timeStamps: array, corresponding times of flight in s.
        plot: bool, to plot :math:`v^2` vs :math:`TOF^2`. Default:True.

    Returns:
        a tuple, (temperature in K along x, temperature in K along y)
    '''
    params = rcParams().params
    factor = (params['pixelSize'] * params['binning'] / params['magnification']) ** 2
    vxSqFit, intx = np.polyfit(timeStamps ** 2, factor * (sizes[:, 0] ** 2), 1)
    vySqFit, inty = np.polyfit(timeStamps ** 2, factor * (sizes[:, 1] ** 2), 1)
    tt = np.linspace(timeStamps[0], timeStamps[-1], 30)
    Tx = params['mass number'] * m_p * vxSqFit / k
    Ty = params['mass number'] * m_p * vySqFit / k
    if plot==True:
        plt.figure(figsize=(8, 3))
        plt.plot(timeStamps[:] ** 2 / (10 ** -6), sizes[:, 0] ** 2 * factor / (10 ** -12), 'o')
        plt.plot(timeStamps[:] ** 2 / (10 ** -6), sizes[:, 1] ** 2 * factor / (10 ** -12), 'o')
        plt.plot(tt ** 2 / (10 ** -6), ((vxSqFit) * (tt) ** 2 + intx) / (10 ** -12), color='C0', ls='-', label='$T_x = {:.3f}\mu K$'.format(Tx/1e-6))
        plt.plot(tt ** 2 / (10 ** -6), ((vySqFit) * (tt) ** 2 + inty) / (10 ** -12), color='C1', ls='-', label='$T_y = {:.3f}\mu K$'.format(Ty/1e-6))
        plt.xlabel('$(TOF (ms))^2$')
        plt.ylabel('$(\sigma (\mu m))^2$')
        plt.legend()
        plt.tight_layout()
    return Tx, Ty


import numpy as np
import matplotlib.pyplot as plt


class GaussianBeam(object):
    def __init__(self, wLen, w0, n=1):
        self.wLen = wLen/n
        self.w0 = w0
        self.zR = np.pi*w0**2/wLen

    def w(self, z):
        '''Waist of the beam at z'''
        return self.w0*np.sqrt((1+(z/self.zR)**2))

    def R(self, z):
        '''Radius of curvature of the wavefront at z'''
        return z*np.sqrt(1+(z/self.zR)**2)

    def gouyPhase(self, z):
        '''Gouy phase at z'''
        return np.arctan2(z, self.zR)

    def field(self, x, y, z):
        '''Complex field value at (x, y, z)'''
        rsq = x**2 + y**2
        k = 2*np.pi/self.wLen
        amp = (self.w0/self.w(z))*np.exp(-rsq/self.w(z)**2)
        phase = k*(z+rsq/(2*self.R(z)))-self.gouyPhase(z)
        return amp*np.exp(-1j*phase)

    def intensity(self, x, y, z):
        '''Intensity at (x, y, z)'''
        rsq = x**2+y**2
        inten = (self.w0/self.w(z))**2*np.exp(-2*rsq/self.w(z)**2)
        return inten

    def complexBeamParameter(self, z):
        '''complex beam parameter q'''
        q = z+1j*self.zR
        return q





from scipy.constants import *

def sigmaGeneral(delta, wLen, Gamma, s=0):
    """
        Calculates the scattering cross-section of a two-level atom with a weak probe.

        Parameters:
            delta: float, detuning of the probe w.r.t the reference in MHz.
            wLen: float, wavelength of the probe in nm
            Gamma: flaot, linewidth of the excited state in MHz.
            s: float, saturation parameter of the probe, :math:`I/I_s`.
        Returns:
            A float, Scattering cross-section.
        Comment:
            All frequencies are in MHz.
    """
    sigma_0 = 3*(wLen*nano)**2/(2*pi)
    return sigma_0/(4*((delta/Gamma)**2+1+s))

def sigmaBlue(delta, A, s=0):
    """
    Calculates the scattering cross-section of Sr for :math:`^1S_0 \\rightarrow ^1P_1` taking into
    account the isotope shift w.r.t. A=88 isotope and hyperfine levels present
    in the excited state.

    Parameters:
        delta: float, detuning of the probe w.r.t the reference in MHz.
        s: float, saturation parameter of the probe, :math:`I/I_s`.
        A: int, mass number of the isotope.
    Returns:
        A float, Scattering cross-section.
    Comment:
        All frequencies are in MHz.
    """
    wLen = 461*nano
    Gamma = 2*pi*32
    sigma_0 = 3*wLen**2/(2*pi)
    calib = 36
    if A == 88:
        return sigma_0/(4*(2*pi*delta/Gamma)**2+1+s)
    elif A == 87:
        zeemanShift = 1.60*calib # shift in spectroscopy in choosing 87^Sr
        delta9half = delta - (zeemanShift-69)
        delta7half = delta - (zeemanShift-9.7)
        delta11half = delta - (zeemanShift-51.8)
        coeff7half = 8/(4*(2*pi*delta7half/Gamma)**2+1+s)
        coeff9half = 10/(4*(2*pi*delta9half/Gamma)**2+1+s)
        coeff11half = 12/(4*(2*pi*delta11half/Gamma)**2+1+s)
        return sigma_0*(0.1826**2)*(coeff7half+coeff9half+coeff11half)
        # the 0.1826 comes from the branching ratio for different hyperfine transitions.
    elif A == 86:
        raise ValueError('The calculation for 86^Sr is not done. Please do it.')
    elif A == 84:
        raise ValueError('The calculation for 84^Sr is not done. Please do it.')


def sigmaRed(delta, s=0):
    """
    Calculates the scattering cross-section of Sr for :math:`^1S_0 \\rightarrow ^3P_1` taking saturation
    into account.

    Parameters:
        delta: float, detuning of the probe w.r.t the reference in kHz.
        s: float, saturation parameter of the probe, :math:`I/I_s`.
    Returns:
        A float, Scattering cross-section.
    Comment:
        All frequencies are in kHz.
    """
    wLen = 689*nano
    Gamma = 7.5
    sigma_0 = 3*wLen**2/(2*pi)
    return sigma_0/(4*(delta/Gamma)**2+1+s)



import numpy as np
from scipy.optimize import curve_fit
from scipy.constants import *
import matplotlib.pyplot as plt
import matplotlib as mpl
from IPython.display import display_latex

def gaussian(x, amplitude, x_0, sigma_x, offset):
    '''
    1D Gaussian function, :math:`f(x) = e^{\\frac{-(x-x_0)^2}{2\sigma_x^2}}+offset`

    Parameters:
        x: 1D array, x variable of the function
        amplitude: float, amplitude of the gaussian
        xo: float, center of the gaussian
        sigma_x: float, width of the gaussian
        offset: float, constant offset.

    Returns:
        1D array, returns f(x)
    '''
    g = offset + amplitude*np.exp(-(x-x_0)**2/(2*sigma_x**2))
    return g

def gaussianFit(x, array, p0=[], bounds=[(), ()], plot=True, display=True):
    """
    Fits the given array to an 1D-gaussian.

    Parameters:
        x: 1darray, the argument values of the gaussian
        array: 1darray, the data to fit to the gaussian
        p0: ndarray, initial guess for the fit params in the form of
            [amplitude, xo, sigma, offset]. Default is None.
        bounds: tuple of lower bound and upper bound for the fit.
            Default is None.
        plot: bool, Default is True
        display: bool, Default is True. Displays the optimal parameters of the fit
    Returns:
        pOpt: optimized parameters in the same order as p0
        pCov: covarience parameters of the fit.
        Read scipy.optimize.curve_fit for details.
    """
    pOpt, pCov = curve_fit(gaussian, x, array, p0=p0, bounds=bounds)
    if display == True:
        labels = ['$A_0$', '$x_0$', '$\sigma_x$', 'offset']
        for i in range(len(labels)):
            display_latex('{} = {:.3f}'.format(labels[i], pOpt[i]), raw=True)
    if plot==True:
        xsmooth = np.linspace(x[0], x[-1], 1000)
        plt.figure(figsize=(5,3.2))
        plt.plot(x, array, 'o', label='data')
        plt.plot(xsmooth, gaussian(xsmooth, *pOpt), '--k', label='fit')
        plt.xlabel('x')
        plt.legend()
        plt.tight_layout()
    return pOpt, pCov

def gaussian2D(X, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    """
    2D Gaussian function

    Parameters:
        X: np.meshgrid,
        amplitude: float, amplitude
        xo: float, x-center
        yo: float, y-center
        sigma_x: float, :math:`\sigma_1`
        sigma_y: float, :math:`\sigma_2`
        theta: float, angle of tilt
        offset: float, offset

    Returns:
        1d array, flattened 2D array. Reshape it like (X[0], X[0][0]) to get a matrix representation of the gaussian.
    """
    x = X[0]
    y = X[1]
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = np.sin(2*theta)*(-1/(2*sigma_x**2) + 1/(2*sigma_y**2))
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = offset + amplitude*np.exp(-(a*((x-xo)**2) + b*(x-xo)*(y-yo) + c*((y-yo)**2)))
    return g.ravel()

def gaussian2DFit(image, p0=None, bounds=[(), ()], plot=True, title=''):
    """
    Fits an image with a 2D gaussian.

    Parameters:
        image: numpy ndarray
        p0: ndarray, initial guess for the fit params in the form of
            [amplitude, xo, yo, sigma_x, sigma_y, theta, offset].
            Default None (fits for OD images).
        bounds: tuple of lower bound and upper bound for the fit.
            Default None (fits for OD images)
        plot:  bool to show the plot of the fit. Default True.
    Returns:
        pOpt: optimized parameters in the same order as p0
        pCov: covarience parameters of the fit.
        Read scipy.optimize.curve_fit for details.
    """
    Ny, Nx = image.shape
    x = np.linspace(0, Nx, Nx, endpoint=False)
    y = np.linspace(0, Ny, Ny, endpoint=False)
    X = np.meshgrid(x, y)
    if (p0==None and bounds==[(),()]):
        p0 = [0.5, Nx/2, Ny/2, Nx/4, Ny/4, 0, 0]
        bounds = ([-0.5, 0.4*Nx, 0.4*Ny, 0.0*Nx, 0.0*Ny, -0.1, -0.5],\
             [10, 0.6*Nx, 0.6*Ny, 0.7*Nx, 0.7*Ny, 0.1, 1])
    else:
        p0 = p0
        bounds = bounds
    pOpt, pCov = curve_fit(gaussian2D, X, image.reshape((Nx*Ny)), p0, bounds=bounds)
    fit = gaussian2D(X, *pOpt).reshape(Ny, Nx)
    if plot==True:
        f, ax = plt.subplots(nrows=1, ncols=5, gridspec_kw={'width_ratios': [4,4,4,4,0.2]}, figsize=(16, 4))
        if int(pOpt[2])>=len(y):
            ax[0].plot(image[-1], 'r.')
        elif int(pOpt[2])<0:
            ax[0].plot(image[0], 'r.')
        else:
            ax[0].plot(image[int(pOpt[2]), :], 'r.')
        ax[0].plot(x, gaussian2D(np.meshgrid(x,int(pOpt[2])), *pOpt), 'k')
        ax[0].set_xlabel('x (pixels)')
        if int(pOpt[1])>=len(x):
            ax[1].plot(image[:, -1], 'r.')
        elif int(pOpt[1])<0:
            ax[1].plot(image[:, 0], 'r.')
        else:
            ax[1].plot(image[:, int(pOpt[1])], 'r.')
        ax[1].plot(y, gaussian2D(np.meshgrid(int(pOpt[1]),y), *pOpt), 'k')
        ax[1].set_xlabel('y (pixels)')
        if pOpt[0]<0.2:
            ax[0].set_ylim(pOpt[6]-pOpt[0]*1.2, pOpt[6]+pOpt[0]*1.2)
            ax[1].set_ylim(pOpt[6]-pOpt[0]*1.2, pOpt[6]+pOpt[0]*1.2)
        ax[2].contour(X[0], X[1], fit, vmin=pOpt[6]-pOpt[0]*0.12, vmax=pOpt[-1]+abs(pOpt[0])*1.2)
        matrix=ax[2].imshow(image, vmin=pOpt[6]-pOpt[0]*0.12, vmax=pOpt[-1]+abs(pOpt[0])*1.2)
        ax[2].set_xlabel('x (pixels)')
        ax[2].set_ylabel('y (pixels)')
        ax[2].grid(False)
        ax[3].imshow(image-fit, vmin=pOpt[6]-pOpt[0]*0.12, vmax=pOpt[-1]+abs(pOpt[0])*1.2)
        ax[3].set_xlabel('x (pixels)')
        ax[3].set_ylabel('y (pixels)')
        ax[3].grid(False)
        f.colorbar(matrix, cax=ax[4])
        f.suptitle(str(title))
        plt.tight_layout()
    return pOpt, pCov


def threeGaussian2D(X, *args):
    amplitudes, xos, yos, sigma_xs, sigma_ys, thetas, offsets = np.array(args).reshape((7, 3))
    x = X[0]
    y = X[1]
    g = 0
    for i in range(len(amplitudes)):
        g += gaussian2D(X, amplitudes[i], xos[i], yos[i], sigma_xs[i], sigma_ys[i], thetas[i], offsets[i])
    return g.ravel()


def threeGaussian2DFit(image, p0, bounds, TOF, plot=True, cropSize=6, logNorm=False):
    Ny, Nx = image.shape
    x = np.linspace(0, Nx, Nx, endpoint=False)
    y = np.linspace(0, Ny, Ny, endpoint=False)
    X = np.meshgrid(x, y)
    # first Gaussian or thick Gaussian should be well fit first. p0 and bounds should take care of this.
    croppedImage = image[p0[2]-cropSize:p0[2]+cropSize, p0[1]-cropSize:p0[1]+cropSize]
    Xc = np.meshgrid(np.arange(0, 2*cropSize), np.arange(0, 2*cropSize))
    croppedP0 = (p0[0], cropSize, cropSize, p0[3], p0[4], p0[5], p0[6])
    b = bounds
    b[0][1], b[0][2] = 0, 0
    b[1][1], b[1][2] = 3*cropSize/2, 3*cropSize/2
    pOpt, pCov = gaussian2DFit(croppedImage, p0=croppedP0,bounds=b, plot=False)
    pOpt[1] = p0[1]-cropSize+pOpt[1]
    pOpt[2] = p0[2]-cropSize+pOpt[2]
    disQuanta = hbar*(2*pi/(689*nano*87*m_p))*TOF*milli
    dis = disQuanta*2.5/(16*micro)
    p0 = [[pOpt[0], pOpt[0]*0.4, pOpt[0]*0.3],
          [pOpt[1], pOpt[1]-dis, pOpt[1]],
          [pOpt[2], pOpt[2]-dis, pOpt[2]-2*dis],
          [pOpt[3], pOpt[3], pOpt[3]],
          [pOpt[4], pOpt[4], pOpt[4]],
          [pOpt[5], pOpt[5], pOpt[5]],
          [pOpt[6], pOpt[6], pOpt[6]]]
    bounds = ([[pOpt[0]*0.0, pOpt[0]*0.0, pOpt[0]*0.0],
               [pOpt[1]-0.1, pOpt[1]-dis*1.3, pOpt[1]*0.7],
               [pOpt[2]-0.1, pOpt[2]-dis*1.3, pOpt[2]-2*dis*1.3],
               [pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8],
               [pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8],
               [pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1],
               [pOpt[6]-150, pOpt[6]-150, pOpt[6]-150]],
              [[pOpt[0]*1.5, pOpt[0]*1.5, pOpt[0]*1.5],
               [pOpt[1]+0.1, pOpt[1]-dis*0.7, pOpt[1]*1.3],
               [pOpt[2]+0.1, pOpt[2]-dis*0.7, pOpt[2]-2*dis*0.7],
               [pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2],
               [pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2],
               [pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2],
               [pOpt[6]+150, pOpt[6]+150, pOpt[6]+150]])
    pOpt, pCov = curve_fit(threeGaussian2D, X, image.reshape((Nx*Ny)), p0=np.array(p0).reshape((21)), bounds=np.array(bounds).reshape((2, 21)))
    fit = threeGaussian2D(X, *pOpt).reshape(Ny, Nx)
    if plot == True:
        f, ax = plt.subplots(nrows=1, ncols=4, gridspec_kw={'width_ratios': [4, 0.2, 4, 0.2]}, figsize=(10, 4))
        if logNorm==True:
            ax[0].contour(X[0], X[1], fit, norm=mpl.colors.LogNorm(vmin=1e-1, vmax=1e-1+abs(pOpt[0])*1.2))
            matrix = ax[0].imshow(image, aspect='auto', norm=mpl.colors.LogNorm(vmin=1e-1, vmax=9e-1+abs(pOpt[0])*1.2))
            residue = ax[2].imshow(image-fit, aspect='auto')
            ax[2].set_title('Residue:'+str(np.round(np.sum(np.sum(image-fit)), 5)))
        else:
            ax[0].contour(X[0], X[1], fit, vmin=pOpt[-1], vmax=pOpt[-1]+abs(pOpt[0])*1.2)
            matrix = ax[0].imshow(image, aspect='auto', vmin=pOpt[-1], vmax=pOpt[-1]+abs(pOpt[0])*1.2)
            residue = ax[2].imshow(image-fit, aspect='auto')
        ax[0].set_xlabel('x (pixels)')
        ax[0].set_ylabel('y (pixels)')
        ax[0].grid(False)
        ax[2].set_xlabel('x(pixels)')
        ax[2].set_ylabel('y(pixels)')
        ax[2].grid(False)
        f.colorbar(matrix, cax=ax[1])
        f.colorbar(residue, cax=ax[3])
        plt.tight_layout()
    return pOpt, pCov


def multipleGaussian2D(X, *args):
    amplitudes, xos, yos, sigma_xs, sigma_ys, thetas, offsets = np.array(args).reshape((7,5))
    x = X[0]
    y = X[1]
    g = 0
    for i in range(len(amplitudes)):
        g += gaussian2D(X, amplitudes[i], xos[i], yos[i], sigma_xs[i], sigma_ys[i], thetas[i], offsets[i])
    return g.ravel()

def multipleGaussian2DFit(image, p0, bounds, TOF, plot=True, cropSize=6, tolerence=0.3, logNorm=False):
    Ny, Nx = image.shape
    x = np.linspace(0, Nx, Nx, endpoint=False)
    y = np.linspace(0, Ny, Ny, endpoint=False)
    X = np.meshgrid(x, y)
    # first Gaussian or thick Gaussian should be well fit first. p0 and bounds should take care of this.
    croppedImage = image[int(p0[2])-cropSize:int(p0[2])+cropSize, int(p0[1])-cropSize:int(p0[1])+cropSize]
    Xc = np.meshgrid(np.arange(0, 2*cropSize), np.arange(0, 2*cropSize))
    croppedP0 = (p0[0], cropSize, cropSize, p0[3], p0[4], p0[5], p0[6])
    b = bounds
    b[0][1], b[0][2] = 0, 0
    b[1][1], b[1][2] = 3*cropSize/2, 3*cropSize/2
    pOpt, pCov = gaussian2DFit(croppedImage, p0=croppedP0,bounds=b, plot=False)
    pOpt[1] = p0[1]-cropSize+pOpt[1]
    pOpt[2] = p0[2]-cropSize+pOpt[2]
    disQuanta = hbar*(2*pi/(689*nano*87*m_p))*TOF*milli
    dis = disQuanta*2.5/(16*micro)
    d = dis
    '''
    p0 = [[pOpt[0], pOpt[0]*0.4, pOpt[0]*0.3, pOpt[0]*0.02, pOpt[0]*0.02, pOpt[0]*0.02, pOpt[0]*0.02, pOpt[0]*0.02],
          [pOpt[1], pOpt[1]-d, pOpt[1], pOpt[1]-d, pOpt[1], pOpt[1]+d, pOpt[1]+d, pOpt[1]-d],
          [pOpt[2], pOpt[2]-d, pOpt[2]-2*d, pOpt[2]-3*d, pOpt[2]-4*d, pOpt[2]-d, pOpt[2]+d, pOpt[2]+d],
          [pOpt[3], pOpt[3], pOpt[3], pOpt[3], pOpt[3], pOpt[3], pOpt[3], pOpt[3]],
          [pOpt[4], pOpt[4], pOpt[4], pOpt[4], pOpt[4], pOpt[4], pOpt[4], pOpt[4]],
          [pOpt[5], pOpt[5], pOpt[5], pOpt[5], pOpt[5], pOpt[5], pOpt[5], pOpt[5]],
          [pOpt[6], pOpt[6], pOpt[6], pOpt[6], pOpt[6], pOpt[6], pOpt[6], pOpt[6]]]
    x = tolerence
    ld = (1-x)*d
    ud = (1+x)*d
    bounds = ([[-pOpt[0]*0.1, -pOpt[0]*0.1, -pOpt[0]*0.1, -pOpt[0]*0.1, -pOpt[0]*0.1, -pOpt[0]*0.1, -pOpt[0]*0.1, -pOpt[0]*0.1],
               [pOpt[1]-x*d, pOpt[1]-ud, pOpt[1]-x*d, pOpt[1]-ud, pOpt[1]-x*d, pOpt[1]+ld, pOpt[1]+ld, pOpt[1]-ud],
               [pOpt[2]-x*d, pOpt[2]-ud, pOpt[2]-2*ud, pOpt[2]-3*ud, pOpt[2]-4*ud, pOpt[2]-ud, pOpt[2]+ld, pOpt[2]+ld],
               [pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8, pOpt[3]*0.8],
               [pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8, pOpt[4]*0.8],
               [pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1, pOpt[5]-0.1],
               [pOpt[6]-150, pOpt[6]-150, pOpt[6]-150, pOpt[6]-150, pOpt[6]-150, pOpt[6]-150, pOpt[6]-150, pOpt[6]-150]],
              [[pOpt[0]*1.5, pOpt[0]*1.5, pOpt[0]*1.5, pOpt[0]*0.7, pOpt[0]*0.7, pOpt[0]*0.5, pOpt[0]*0.5, pOpt[0]*0.5],
               [pOpt[1]+x*d, pOpt[1]-ld, pOpt[1]+x*d, pOpt[1]-ld, pOpt[1]+x*d, pOpt[1]+ud, pOpt[1]+ud, pOpt[1]-ld],
               [pOpt[2]+x*d, pOpt[2]-ld, pOpt[2]-2*ld, pOpt[2]-3*ld, pOpt[2]-4*ld, pOpt[2]-ld, pOpt[2]+ud, pOpt[2]+ud],
               [pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2, pOpt[3]*1.2],
               [pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2, pOpt[4]*1.2],
               [pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2, pOpt[5]+0.2],
               [pOpt[6]+150, pOpt[6]+150, pOpt[6]+150, pOpt[6]+150, pOpt[6]+150, pOpt[6]+150, pOpt[6]+150, pOpt[6]+150]])
    '''
    p0 = [[pOpt[0], pOpt[0] * 0.4, pOpt[0] * 0.3, pOpt[0] * 0.02, pOpt[0] * 0.02],
          [pOpt[1], pOpt[1] - dis, pOpt[1], pOpt[1] - dis, pOpt[1]],
          [pOpt[2], pOpt[2] - dis, pOpt[2] - 2 * dis, pOpt[2] - 3 * dis, pOpt[2] - 4 * dis],
          [pOpt[3], pOpt[3], pOpt[3], pOpt[3], pOpt[3]],
          [pOpt[4], pOpt[4], pOpt[4], pOpt[4], pOpt[4]],
          [pOpt[5], pOpt[5], pOpt[5], pOpt[5], pOpt[5]],
          [pOpt[6], pOpt[6], pOpt[6], pOpt[6], pOpt[6]]]
    bounds = (
    [[pOpt[0] * -0.1, pOpt[0] * -0.1, pOpt[0] * -0.1, pOpt[0] * -0.1, pOpt[0] * -0.1],
     [pOpt[1] - 0.1*dis, pOpt[1] - dis * 1.1, pOpt[1] - 0.1 * dis, pOpt[1] - dis * 1.1, pOpt[1] - 0.1*dis],
     [pOpt[2] - 0.1*dis, pOpt[2] - dis * 1.1, pOpt[2] - 2 * dis * 1.1, pOpt[2] - 3 * dis * 1.1, pOpt[2] - 4 * dis * 1.1],
     [pOpt[3] * 0.9, pOpt[3] * 0.9, pOpt[3] * 0.9, pOpt[3] * 0.9, pOpt[3] * 0.9],
     [pOpt[4] * 0.9, pOpt[4] * 0.9, pOpt[4] * 0.9, pOpt[4] * 0.9, pOpt[4] * 0.9],
     [pOpt[5] - 0.1, pOpt[5] - 0.1, pOpt[5] - 0.1, pOpt[5] - 0.1, pOpt[5] - 0.1],
     [pOpt[6] - 150, pOpt[6] - 150, pOpt[6] - 150, pOpt[6] - 150, pOpt[6] - 150]],
    [[pOpt[0] * 1.5, pOpt[0] * 1.5, pOpt[0] * 1.5, pOpt[0] * 0.7, pOpt[0] * 0.7],
     [pOpt[1] + 0.1*dis, pOpt[1] - dis * 0.9, pOpt[1] * 1.1, pOpt[1] - dis * 0.9, pOpt[1] + 0.1*dis],
     [pOpt[2] + 0.1*dis, pOpt[2] - dis * 0.9, pOpt[2] - 2 * dis * 0.9, pOpt[2] - 3 * dis * 0.9, pOpt[2] - 4 * dis * 0.9],
     [pOpt[3] * 1.1, pOpt[3] * 1.1, pOpt[3] * 1.1, pOpt[3] * 1.1, pOpt[3] * 1.1],
     [pOpt[4] * 1.1, pOpt[4] * 1.1, pOpt[4] * 1.1, pOpt[4] * 1.1, pOpt[4] * 1.1],
     [pOpt[5] + 0.2, pOpt[5] + 0.2, pOpt[5] + 0.2, pOpt[5] + 0.2, pOpt[5] + 0.2],
     [pOpt[6] + 150, pOpt[6] + 150, pOpt[6] + 150, pOpt[6] + 150, pOpt[6] + 150]])#'''
    pOpt, pCov = curve_fit(multipleGaussian2D, X, image.reshape((Nx*Ny)), p0=np.array(p0).reshape((35)), bounds=np.array(bounds).reshape((2, 35)))
    fit = multipleGaussian2D(X, *pOpt).reshape(Ny, Nx)
    if plot == True:
        f, ax = plt.subplots(nrows=1, ncols=4, gridspec_kw={'width_ratios': [4, 0.2, 4, 0.2]}, figsize=(10, 4))
        if logNorm==True:
            ax[0].contour(X[0], X[1], fit, norm=mpl.colors.LogNorm(vmin=1e-1, vmax=1e-1+abs(pOpt[0])*1.2))
            matrix = ax[0].imshow(image, aspect='auto', norm=mpl.colors.LogNorm(vmin=1e-1, vmax=9e-1+abs(pOpt[0])*1.2))
            residue = ax[2].imshow(image-fit, aspect='auto')
            ax[2].set_title('Residue:'+str(np.round(np.sum(np.sum(image-fit)), 5)))
        else:
            ax[0].contour(X[0], X[1], fit, vmin=pOpt[-1], vmax=pOpt[-1]+abs(pOpt[0])*1.2)
            matrix = ax[0].imshow(image, aspect='auto', vmin=pOpt[-1], vmax=pOpt[-1]+abs(pOpt[0])*1.2)
            residue = ax[2].imshow(image-fit, aspect='auto')
        ax[0].set_xlabel('x (pixels)')
        ax[0].set_ylabel('y (pixels)')
        ax[0].grid(False)
        ax[2].set_xlabel('x(pixels)')
        ax[2].set_ylabel('y(pixels)')
        ax[2].grid(False)
        f.colorbar(matrix, cax=ax[1])
        f.colorbar(residue, cax=ax[3])
        plt.tight_layout()
    return pOpt, pCov

def lorentzian(x, amplitude, x_0, gamma, offset):
    l = offset + amplitude*(gamma/2)/((x-x_0)**2+(gamma/2)**2)
    return l

def lorentzianFit(x, array, p0=[], bounds=[(), ()], plot=True, display=True):
    """
    Fits the given array to a Lorentzian.

    Parameters:
        array: 1darray, the data to fit to the lorentzian
        p0: ndarray, initial guess for the fit params in the form of
            [amplitude, xo, gamma, offset]. Default is [].
        bounds: tuple of lower bound and upper bound for the fit.
            Default is [(), ()].
        plot: bool, Default is True
        display: bool, Default is True. Displays the optimal parameters of the fit
    Returns:
        pOpt: optimized parameters in the same order as p0
        pCov: covarience parameters of the fit.
        Read scipy.optimize.curve_fit for details.
    """
    pOpt, pCov = curve_fit(lorentzian, x, array, p0, bounds=bounds)
    if display == True:
        labels = ['$A_0$', '$x_0$', '$\gamma$', 'offset']
        for i in range(len(labels)):
            display_latex('{} = {:.3f}'.format(labels[i], pOpt[i]), raw=True)
    if plot == True:
        xsmooth = np.linspace(x[0], x[-1], 1000)
        plt.figure(figsize=(5, 3.2))
        plt.plot(x, array, 'o', label='data')
        plt.plot(xsmooth, lorentzian(xsmooth, *pOpt), '--k', label='fit')
        plt.xlabel('x')
        plt.legend()
        plt.tight_layout()
    return pOpt, pCov




from IPython.display import clear_output
from IPython.display import display_latex
from sympy import latex

def update_progress(progress, clear=True, comment=''):
    '''
    Progress bar generated in a loop.

    Parameters:
        progress: int, current index
        clear: bool, to clear the last output of the cell. Default True
        comment: string, description of the progress bar.

    Returns:
        None
    '''
    bar_length = 50
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
    if progress < 0:
        progress = 0
    if progress >= 1:
        progress = 1
    block = int(round(bar_length * progress))
    if clear:
        clear_output(wait = True)
    text = 'Progress ' + str(comment) + ': [{0}] {1:.1f}%'.format("#" * block + "-" * (bar_length - block),
                                                                  progress * 100)
    print(text)

def disp(idx, symObj):
    '''
    Displays sympy symbolic objects in latex as an equation.

    Parameters:
        idx: string, latex code of varaible in RHS.
        symObj: sympy.symbol object, symbol object to be displayed in LHS.

    Returns:
        None.
    '''
    eqn = '\\[' + idx + ' = ' + latex(symObj) + '\\]'
    display_latex(eqn,raw=True)
    return


import os
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from warnings import warn
import json
from importlib import resources
import io
from scipy.constants import *
import sys
import datetime as dt
from sif_parser import np_open

class rcParams():
    """
    A class that is designed to read and update resource parameters of the package. These parameters are usually the
    hardware parameters that are used while converting certain digital values to real units. Currently these are
    imaging parameters like magnification, pixel size etc. The idea behind this is that once these parameters are set,
    these are used by many function and classes throughout the package without having to explicitly pass them. Open
    the file rcParams.json using the notepad to see the current parameters used throughout the library.

    Attributes:

        params: dict, current parameters used by the package
    """
    def __init__(self):
        with resources.open_binary("ColdTOFU", "rcParams.json") as file:
            content = file.read()
            self.path = os.path.abspath(file.name)
            self.params = json.load(io.BytesIO(content))
            file.close()
    def update(self, key, value):
        """
        A method to update the rcParams.

        Args:

            key: string, key of the parameter. Careful about the spelling of the key. If its wrong,
                then you will add another parameter with the wrong spelling instead of updating the
                desired parameter.
            value: value of the parameter to update.
        """
        with open(self.path, 'w') as file:
            self.params[key] = value
            json.dump(self.params, file)
            file.close()


class ShadowImage(object):
    """
    The main class to extract relevant information from the shadow imaging
    sequences that are obtained in the experiment. The multiple image sequence
    consists of the shadow or the absorption image, image of the incident probe
    and the image of the background for as many runs of the experiment.
    
    Parameters:
        filePath: str, Path to the multiple-image file

    Attributes:

        filePath: str, Path to the multiple-image file
        ext: str, extension of the image file passed.
        im: PIL.TiffImagePlugin.TiffImageFile or AndorSifFile._SifFrames
        n: int, total number of data = number of images/3
        frames: ndarray, all the frames in the image file
        tags: dict, the named tag dictionary of the TiffImage or properties of sif file
        transmission: ndarray, All the transmission images after
                       subtracting the background
        incidence: ndarray, All the images of incident probe after
                    subtracting the background
        OD: ndarray, optical depth all the runs of the experiment calculated from
             the transmission and incidence
        ODaveraged: ndarray, OD averaged with averaging like a superloop
        averagedOD: ndarray, OD of the averaged signal with averaging on superloop
        averagedOD2: ndarray, OD of the averaged signal with averaging on loop
    """
    def __init__(self, filePath):
        self.filePath = filePath
        self.ext = os.path.splitext(filePath)[1]
        if self.ext == '.tif':
            self.im = Image.open(filePath, formats=['TIFF'])
            self.tags = self.im.tag.named()
        elif self.ext == '.sif':
            if sys.platform.startswith('win'):
                from .AndorSifReader import AndorSifFile
                self.im = AndorSifFile(filePath).signal
                self.tags = self.im.props
                self.frames = self.im.data
            elif sys.platform.startswith('lin'):
                try:
                    data_read = np_open(filePath)
                    self.frames = data_read[0]
                    self.im = Image.fromarray(data_read[0][0])
                    self.im.n_frames = len(self.frames)
                    self.tags = data_read[1]
                except IOError:
                    raise IOError('Sif file could not be read or file doesn\'t exist. Try on Windows machine.')
            elif sys.platform.startswith('dar'):
                try:
                    data_read = np_open(filePath)
                    self.frames = data_read[0]
                    self.im = Image.fromarray(data_read[0][0])
                    self.im.n_frames = len(self.frames)
                    self.tags = data_read[1]
                except IOError:
                    raise IOError('Sif file could not be read or file doesn\'t exist. Try on Windows machine.')
            else:
                warn('Andor *.sif files could not be read in your OS as ATSIFIO64.dll is only available for windows.')
        else:
            raise IOError('Invalid file!')
        if self.im.n_frames%2!=0:
            warn('Not a valid fluorescence image. \
                  No. of images in the file is not a multiple of 2.')
        self.n = self.im.n_frames//3
        self.frames = self.images()
        self.params = rcParams().params
        self.docs = None

    def images(self):
        """
        Returns the images present in the ShadowImage object as an ndarray.
        """
        if self.ext == '.tif':
            frames = np.zeros((self.im.n_frames, self.im.height, self.im.width))
            for i in range(self.im.n_frames):
                self.im.seek(i)
                frames[i] = np.array(self.im)
            return frames
        elif self.ext == '.sif' and sys.platform.startswith('win'):
            frames = self.im.data
            return frames
        else:
            return self.frames

    def opticalDepth(self, xSpan, ySpan):
        """
        Returns the optical depth calculated from the ShadowImage as an ndarray.

        Parameters:
            xSpan: list, containing the range of pixels of the image in x direction.
            ySpan: list, containing the range of pixels of the image in y direction.
        """
        self.transmission = np.zeros((self.n, self.im.height, self.im.width))
        self.incidence = np.zeros((self.n, self.im.height, self.im.width))
        for i in range(self.n):
            self.transmission[i] = self.frames[3*i]-self.frames[3*i+2]
            self.incidence[i] = self.frames[3*i+1]-self.frames[3*i+2]
        self.incidence[self.incidence == 0] = 1e-5
        self.incidence[self.incidence != self.incidence] = 1e-5
        self.transmission[self.transmission == 0] = 1e-5
        self.transmission[self.transmission != self.transmission] = 1e-5
        T = self.transmission[:,ySpan[0]:ySpan[1], xSpan[0]:xSpan[1]]\
                               /self.incidence[:,ySpan[0]:ySpan[1], xSpan[0]:xSpan[1]]
        T[T <= 0] = 1e-20
        T[T != T] = 1e-20
        #T[T > 1+1e-3] = 1
        self.OD = -np.log(T)
        return self.OD

    def opticalDepthAveraged(self, nAveraging):
        """
        Calculates the optical depth from every triad of the images and
        the take average with nAveraging as the Superloop in the experiment.

        Parameters:
            nAveraging: int, number of averaging to be used.

        Returns:
            an ndarray of length equal to Serie in the experiment.
        """
        self.nAveraging = nAveraging
        self.nSamples = int(self.n/nAveraging)
        result = np.zeros((self.nSamples, self.im.height, self.im.width))
        ShadowImage.opticalDepth(self, [0, self.im.width], [0, self.im.height])
        for j in range(self.nSamples):
            for i in range(self.nAveraging):
                result[j] += self.OD[self.nSamples*i+j]
            result[j] /= self.nAveraging
        self.ODaveraged = result
        return self.ODaveraged

    def averagedSignalOD(self, nAveraging, truncate=()):
        """
        Calculates the average signal with nAveraging being the Superloop
        in the experiment and finds the optical depth after the averaging.

        Parameters:
            nAveraging: int, number of averaging to be used.
            truncate: tuple or list, the superloops to be ignored while averaging. Default is ().

        Returns:
            an ndarray of length equal to Serie in the experiment.
        """
        self.nAveraging = nAveraging
        self.nSamples = int(self.n/nAveraging)
        self.averagedTransmission = np.zeros((self.nSamples, self.im.height, self.im.width))
        self.averagedIncidence = np.zeros((self.nSamples, self.im.height, self.im.width))
        ShadowImage.opticalDepth(self, [0, self.im.height], [0, self.im.height])
        for j in range(self.nSamples):
            for i in range(self.nAveraging):
                if not(i in truncate):
                    self.averagedTransmission[j] += self.transmission[self.nSamples*i+j]
                    self.averagedIncidence[j] += self.incidence[self.nSamples*i+j]
        self.averagedTransmission /= (nAveraging-len(truncate))
        self.averagedIncidence /= (nAveraging-len(truncate))
        self.averagedIncidence[self.averagedIncidence == 0] = 1e-5        
        T = self.averagedTransmission/self.averagedIncidence
        T[T != T] = 1e-20
        T[T <= 0] = 1e-20
        self.averagedOD = -np.log(T)
        return self.averagedOD

    def averagedSignalOD2(self, nAveraging, truncate=()):
        """
        Calculates the average signal with nAveraging being the loops
        in the experiment and finds the optical depth after the averaging.

        Parameters:
            nAveraging: int, number of averaging to be used.
            truncate: tuple or list, the superloops to be ignored while averaging. Default is ().

        Returns:
            an ndarray of length equal to Serie in the experiment.
        """
        self.nAveraging = nAveraging
        self.nSamples = int(self.n/nAveraging)
        self.averagedTransmission2 = np.zeros((self.nSamples, self.im.height, self.im.width))
        self.averagedIncidence2 = np.zeros((self.nSamples, self.im.height, self.im.width))
        ShadowImage.opticalDepth(self, [0, self.im.height], [0, self.im.height])
        for j in range(self.nSamples):
            for i in range(self.nAveraging):
                self.averagedTransmission2[j] += self.transmission[self.nSamples*j+i]
                self.averagedIncidence2[j] += self.incidence[self.nSamples*j+i]
        
        self.averagedTransmission /= (nAveraging-len(truncate))
        self.averagedIncidence /= (nAveraging-len(truncate))
        self.averagedIncidence2[self.averagedIncidence == 0] = 1e-5
        T = self.averagedTransmission2/self.averagedIncidence2
        T[T != T] = 1e-20
        T[T <= 0] = 1e-20
        self.averagedOD2 = -np.log(T)
        return self.averagedOD2
       
    def plotAveragedSignalOD(self, nAveraging, ROI, truncate=()):
        """
        Calculates and plots the average signal with nAveraging being the Superloop
        in the experiment and finds the optical depth after the averaging.

        Parameters:
            nAveraging: int, number of averaging to be used.
            ROI: list, ROI in the images to be plotted.
            truncate: tuple or list, the superloops to be ignored while averaging. Default is ().
        Returns None.
        """
        OD = self.averagedSignalOD(nAveraging, truncate)
        f, ax = plt.subplots(nrows=len(OD), ncols=1, figsize=(4,len(OD)*2))
        for i in range(len(OD)):
            a = ax[i].imshow(OD[i, ROI[0]:ROI[1], ROI[2]:ROI[3]])
            ax[i].grid(False)
            f.colorbar(a, ax=ax[i])
        plt.tight_layout()
        return None
        
    def comment(self, comment):
        """
        Adds comment to the tif image under the ImageDescription tag
        and replaces it silently with the new image, if the user has the
        permissions.

        Parameters:
            comment: str, comment to be added.
        """
        path, ext = os.path.splitext(self.filePath)
        newPath = path+'cmtd'+ext
        self.im.save(newPath, save_all=True, description=str(comment))
        self.im.close()
        #os.replace(newPath, self.filePath)

    def description(self):
        """
        Returns:
             the ImageDescription tag of the tiff image file.
        """
        try:
            r = self.tags['ImageDescription']
        except KeyError:
            r = 'No description to show!'
        return r

    def redProbeIntensity(self, ROI):
        """
        Estimates red probe intensity and power from the reference image using the losses in imaging system
        and quantum effeiciency of the camera for red light. rcParams used in calculating the intensity are
        ``quantum efficiency``, ``pixelSize``, ``binning`` and ``magnification``.

        Parameters:
            ROI: list, ROI in which to estimate the intensity.
        Returns:
             intensity (:math:`\mu W/cm^2`) and power (:math:`\mu W`)
        """
        try:
            probeImage = self.averagedIncidence[0][ROI[0]:ROI[1], ROI[2]:ROI[3]]
        except AttributeError:
            print("Call averagedSignalOD to calculate averagedIncidece before calling probeIntensity.")
            return
        imConstant = self.params['binning']*self.params['pixelSize']/self.params['magnification']
        totalCount = np.sum(np.sum(probeImage))
        area = len(probeImage)*len(probeImage[0])*imConstant**2
        if self.ext=='.tif': # for PCO panda 4.2 bi camera red imaging
            photons = totalCount*0.8/(self.params['quantum efficiency'])
            energy = photons*h*c/(689*nano)
            power = energy/(0.95*80*micro) # 0.95 to account for filter and losses on optics
            intensity = (power/1e-6)/(area*10**4) # in micro Watt/cm^2
            return intensity, power/1e-6
        elif self.ext == '.sif': # for andor
            raise NotImplementedError
    
    def blueProbeIntensity(self, ROI):
        """
        Estimates blue probe intensity and power from the reference image using the losses in imaging system
        and quantum effeiciency of the camera for red light. rcParams used in calculating the intensity are
        ``quantum efficiency``, ``pixelSize``, ``binning`` and ``magnification``.

        Parameters:
            ROI: list, ROI in which to estimate the intensity.
        Returns:
             intensity (:math:`\mu W/cm^2`) and power (:math:`\mu W`)
        """
        try:
            probeImage = self.averagedIncidence[0][ROI[0]:ROI[1], ROI[2]:ROI[3]]
        except AttributeError:
            print("Call averagedSignalOD to calculate averagedIncidece before calling probeIntensity.")
            return
        imConstant = self.params['binning']*self.params['pixelSize']/self.params['magnification']
        totalCount = np.sum(np.sum(probeImage))
        area = len(probeImage)*len(probeImage[0])*imConstant**2
        if self.ext=='.tif': # for PCO panda 4.2 bi camera blue imaging
            photons = totalCount*0.8/(self.params['quantum efficiency'])
            energy = photons*h*c/(461*nano)
            power = energy/(0.95*20*micro) # 0.95 to account for filter and losses on optics
            intensity = (power/1e-6)/(10**4*area) # in micro Watt/cm^2
            return intensity, power
        elif self.ext == '.sif': # for andor
            raise NotImplementedError

    def __str__(self):
        return str(self.tags)

    def __add__(self, im2):
        if isinstance(im2, ShadowImage)==True:
            n = im2.im.n_frames//3
            frames = im2.images()
            self.n +=n
            self.frames = np.concatenate((self.frames, frames), axis=0)
            return self
        else:
            warn('add operation types is not same')

    ### Shadow image context manager methods for logging
    def __enter__(self):
        self.call_stack = []
        self.original_trace_function = sys.gettrace()
        sys.settrace(self.log_function_call)
        now = dt.datetime.now()
        filename='log_'+now.strftime('%Y%m%d')+'.md'
        try:
            self.log_file = open('ColdTOFU_results/'+filename, 'a+')
        except FileNotFoundError:
            pwd = os.getcwd()
            os.mkdir(os.path.join(pwd, 'ColdTOFU_results'))
            self.log_file = open('ColdTOFU_results/'+filename, 'a+')
        self.log_file.write('_'*20+'\n')
        self.log_file.write(now.strftime('%Y-%m-%d %H:%M:%S')+'\n')
        self.log_file.write('* path: {}, number of shadow images: {}\n'.format(self.filePath, self.n))
        return self

    def log_function_call(self, frame, event, args):
        if event == "call":
            function_name = frame.f_code.co_name
            module_name = frame.f_globals.get('__name__', '')
            if module_name.startswith('ColdTOFU.spectrum') or module_name.startswith('ColdTOFU.numberOfAtoms'):
                self.call_stack.append(f'`{module_name}.{function_name}`')

    def write_docs(self, docs):
        self.log_file.write('* {}\n'.format(docs))

    def __exit__(self, exc_type, exc_value, exc_tb):
        sys.settrace(self.original_trace_function)
        self.log_file.write('* averaging: {}\n* function calls:\n\t* '.format(self.n//self.nSamples))
        self.log_file.write('\n\t* '.join(self.call_stack))
        self.log_file.write('\n')
        if exc_type!=None:
            self.write_docs(exc_type)
            self.write_docs(exc_value)
        self.log_file.close()

class FluorescenceImage(object):
    """
    The main class to extract relevant information from the fluorescence image
    sequences that are obtained in the experiment. The multiple image sequence
    consists of fluorescence signal image and a background image acquired immediately
    after the signal from the atomic cloud for as many runs of the experiment.
    
    Parameters:
        filePath: str, Path to the multiple-image file

    Attributes:
        filePath: str, Path to the multiple-image file
        im: PIL.TiffImagePlugin.TiffImageFile or AndorSifFile._SifFrames
        n: int, total number of data = number of images/3
        frames: ndarray, all the frames in the image file
        tags: dict, the named tag dictionary of the TiffImage or properties of sif file
    """
    def __init__(self, filePath):
        self.filePath = filePath
        self.ext = os.path.splitext(filePath)[1]
        if self.ext == '.tif':
            self.im = Image.open(filePath)
            self.tags = self.im.tag.named()
        elif self.ext == '.sif':
            if sys.platform.startswith('win'):
                from .AndorSifReader import AndorSifFile
                self.im = AndorSifFile(filePath).signal
                self.tags = self.im.props
                self.frames = self.im.data
            elif sys.platform.startswith('lin'):
                try:
                    data_read = np_open(filePath)
                    self.frames = data_read[0]
                    self.im = Image.fromarray(data_read[0][0])
                    self.im.n_frames = len(self.frames)
                    self.tags = data_read[1]
                except IOError:
                    raise IOError('Sif file could not be read or file doesn\'t exist. Try on Windows machine.')
            elif sys.platform.startswith('dar'):
                try:
                    data_read = np_open(filePath)
                    self.frames = data_read[0]
                    self.im = Image.fromarray(data_read[0][0])
                    self.im.n_frames = len(self.frames)
                    self.tags = data_read[1]
                except IOError:
                    raise IOError('Sif file could not be read or file doesn\'t exist. Try on Windows machine.')
            else:
                warn('Andor *.sif files could not be read in your OS as ATSIFIO64.dll is only available for windows.')
        else:
            raise IOError('Invalid file!')
        if self.im.n_frames%2!=0:
            warn('Not a valid fluorescence image. \
                  No. of images in the file is not a multiple of 2.')
        self.n = self.im.n_frames//2
        self.frames = self.images()

    def images(self):
        """
        Returns the images present in the multiple image file as an ndarray.
        """
        if self.ext == '.tif':
            frames = np.zeros((self.im.n_frames, self.im.height, self.im.width))
            for i in range(self.im.n_frames):
                self.im.seek(i)
                frames[i] = np.array(self.im)
            return frames
        elif self.ext == '.sif' and sys.platform.startswith('win'):
            frames = self.im.data
            return frames
        else:
            return self.frames
    
    def fluorescence(self, xSpan, ySpan):
        """
        Returns the fluorence calculated from the FluorencenceImage as an ndarray.

        Parameters:
            xSpan: a list containing the range of pixels of the image in x direction.
            ySpan: a list containing the range of pixels of the image in y direction.
        """
        self.fluorescence = np.zeros((self.n, self.im.height, self.im.width))
        for i in range(self.n):
            self.fluorescence[i] = self.frames[2*i]-self.frames[2*i+1]

        self.fluorescence[self.fluorescence <= 0] = 1e-5
        self.fluorescence[self.fluorescence != self.fluorescence] = 1000
        return self.fluorescence
    
    def averagedSignal(self, nAveraging, truncate=()):
        """
        Calculates the average signal with nAveraging being the Superloop
        in the experiment and finds the fluorescence after the averaging.

        Parameters:
            nAveraging: :type: int, the repetitions of the experiment
            truncate: :type: tuple or list, the super loops to be ignored.

        Returns:
            an ndarray of length equal to Serie in the experiment.
        """
        self.nAveraging = nAveraging
        self.nSamples = int(self.n/nAveraging)
        self.averagedFluorescence = np.zeros((self.nSamples, self.im.height, self.im.width))
        self.averagedBackground = np.zeros((self.nSamples, self.im.height, self.im.width))
        FluorescenceImage.fluorescence(self, [0, self.im.width], [0, self.im.height])
        for j in range(self.nSamples):
            for i in range(self.nAveraging):
                if not(i in truncate):
                    self.averagedFluorescence[j] += self.fluorescence[self.nSamples*i+j]
                    self.averagedBackground[j] += self.frames[2*(self.nSamples*i+j)+1]
        self.averagedFluorescence[self.averagedFluorescence <= 0] = 1e-5
        self.averagedBackground /= (nAveraging-len(truncate))
        return self.averagedFluorescence/(nAveraging-len(truncate))
    
    def plotAveragedSignal(self, nAveraging, ROI, truncate=()):
        """
        Calculates and plots the average signal with nAveraging being the Superloop
        in the experiment.

        Returns:
             None.
        """
        fl = self.averagedSignal(nAveraging, truncate)
        w = int(self.im.width/(self.im.width+self.im.height)*4)
        h = int(self.im.height/(self.im.width+self.im.height)*4)
        f, ax = plt.subplots(nrows=len(fl), ncols=1, figsize=(w, h))
        for i in range(len(fl)):
            a = ax[i].imshow(fl[i, ROI[0]:ROI[1], ROI[2]:ROI[3]])
            ax[i].grid(False)
            f.colorbar(a, ax=ax[i])
        plt.tight_layout()
        return None

    def __str__(self):
        return str(self.tags)

    def __enter__(self):
        self.call_stack = []
        self.original_trace_function = sys.gettrace()
        sys.settrace(self.log_function_call)
        now = dt.datetime.now()
        filename = 'log_'+now.strftime('%Y%m%d')+'.md'
        try:
            self.log_file = open('ColdTOFU_results/'+filename, 'a+')
        except FileNotFoundError:
            pwd = os.getcwd()
            os.mkdir(os.path.join(pwd, 'ColdTOFU_results'))
            self.log_file = open('ColdTOFU_results/'+filename, 'a+')
        self.log_file.write('_'*20+'\n')
        self.log_file.write(now.strftime('%Y-%m-%d %H:%M:%S')+'\n')
        self.log_file.write('* path: {}, number of fluorescence images: {}\n'.format(self.filePath, self.n))
        return self

    def log_function_call(self, frame, event, args):
        if event == "call":
            function_name = frame.f_code.co_name
            module_name = frame.f_globals.get('__name__', '')
            if module_name.startswith('ColdTOFU.spectrum') or module_name.startswith('ColdTOFU.numberOfAtoms'):
                self.call_stack.append(f'`{module_name}.{function_name}`')

    def write_docs(self, docs):
        self.log_file.write('* {}\n'.format(docs))

    def __exit__(self, exc_type, exc_value, exc_tb):
        sys.settrace(self.original_trace_function)
        self.log_file.write('* averaging: {}\n* function calls:\n\t* '.format(self.n//self.nSamples))
        self.log_file.write('\n\t* '.join(self.call_stack))
        self.log_file.write('\n')
        if exc_type!=None:
            self.write_docs(exc_type)
            self.write_docs(exc_value)
        self.log_file.close()

def approximatePositionOfTheCloud(meanOD):
    '''
    A function to locate the atomic cloud in images.

    Parameters:
        meanOD: 2darray, image data usually OD image
    Returns:
        a tuple, (y center, x center, approximate size)
    '''
    y_, x_ = np.shape(meanOD)
    size = 0
    binary = meanOD > 0.5*(np.max(meanOD)-np.min(meanOD))
    i0, i1 = np.nonzero(binary)
    if len(i0) >= 0.5*y_*x_:
        warn('Cloud center couldn\'t be determined. Images of the spectroscopy are likey to be noisy.'+
             ' Choosing the center of the image as the approximate position of the cloud')
        y = y_//2
        x = x_//2
    elif len(i0) == 0:
        warn('Images likey do not contain any atoms.'+
             ' Choosing the center of the image as the approximate position of the cloud')
        y = y_//2
        x = x_//2
    else:
        y = int(np.mean(i0))
        x = int(np.mean(i1))
        size = np.sqrt(len(i0)/np.pi)
    return y, x, size


def approximateROI(odimages):
    '''
    A function to provide approximate ROIs for fitting atomic cloud in the images using the function
    approximatePositionOfTheCloud.

    Parameters:
        odimages: list of 2darrays(od images) or a 3darray for which ROIs have to be found

    Returns:
        2darray of dimension (odimages.shape[0], 4), [[y_start, y_end, x_start, x_end] for image 1, ...]
    '''
    ROIs = []
    for o in odimages:
        y, x, size = approximatePositionOfTheCloud(o)
        y_, x_ = np.shape(o)
        if size==0:
            warn('ROIs could not be estimated. Set the ROIs manually.')
            size=0.1*(x+y)/2
        y0, y1, x0, x1 = y-2*int(size), y+2*int(size), x-2*int(size), x+2*int(size)
        if y0<0:
            y0=0
        if y1>y_:
            y1=y_-1
        if x0<0:
            x0=0
        if x1>x_:
            x1=x_-1
        ROIs.append([y0,y1,x0,x1])
    res = np.array(ROIs, dtype=int)
    return res


import numpy as np
from scipy.sparse import csr_matrix, kron
from scipy.sparse import identity as Id

def S_m(s):
    '''
    Matrix representation of ladder lowering operator for spin, in general, angular momentum.

    Parameters:
        s: spin quantum number.
    Returns:
        a sparse matrix representation of :math:`S^-` of shape (2s+1, 2s+1) in Dicke basis, {:math:`|s, m_s\\rangle`}.
    '''
    if s<0:
        raise ValueError('spin has to be non-negative')
    else:
        m_s = -np.arange(-s, s+1, 1)
        rows = np.arange(0,2*s,1)
        cols = np.arange(1,2*s+1,1)
        values = np.array([np.sqrt(s*(s+1)-m*(m-1)) for m in m_s[:-1]])
        spar = csr_matrix((values, (rows, cols)), shape=(int(2*s+1), int(2*s+1)))
        return spar

def S_p(s):
    '''
    Matrix representation of ladder raising operator for spin, in general, angular momentum.

    Parameters:
        s: spin quantum number.
    Returns:
        a sparse matrix representation of :math:`S^+` of shape (2s+1, 2s+1) in Dicke basis, {:math:`|s, m_s\\rangle`}.
    '''
    if s<0:
        raise ValueError('spin has to be non-negative')
    else:
        m_s = -np.arange(-s, s+1, 1)
        rows = np.arange(1,2*s+1,1)
        cols = np.arange(0,2*s,1)
        values = np.array([np.sqrt(s*(s+1)-m*(m+1)) for m in m_s[1:]])
        spar = csr_matrix((values, (rows, cols)), shape=(int(2*s+1), int(2*s+1)))
        return spar

def S_x(s):
    '''
    Matrix representation of x component of the operator for spin, in general, angular momentum.

    Parameters:
        s: spin quantum number.
    Returns:
        a sparse matrix representation of :math:`S_x` of shape (2s+1, 2s+1) in Dicke basis, {:math:`|s, m_s\\rangle`}.
    '''
    return (S_p(s)+S_m(s))/2

def S_y(s):
    '''
    Matrix representation of y component of the operator for spin, in general, angular momentum.

    Parameters:
        s: spin quantum number.
    Returns:
        a sparse matrix representation of :math:`S_y` of shape (2s+1, 2s+1) in Dicke basis, {:math:`|s, m_s\\rangle`}.
    '''
    return (S_p(s)-S_m(s))/2j

def S_z(s):
    '''
    Matrix representation of z component of the operator for spin, in general, angular momentum.

    Parameters:
        s: spin quantum number.
    Returns:
        a sparse matrix representation of :math:`S_z` of shape (2s+1, 2s+1) in Dicke basis, {:math:`|s, m_s\\rangle`}.
    '''
    if s<0:
        raise ValueError('spin has to be non-negative')
    else:
        m_s = -np.arange(-s, s+1, 1)
        rows = np.arange(0,2*s+1,1)
        cols = np.arange(0,2*s+1,1)
        values = np.array([m for m in m_s[:]])
        spar = csr_matrix((values, (rows, cols)), shape=(int(2*s+1), int(2*s+1)))
        return spar

def Spin(s):
    '''
    Matrix representation of quantum mechanical spin, in general, angular momentum, s.

    Parameters:
        s: int or half int, the spin quantum number.
    Returns:
        a tuple of sparse matrices corresponding to :math:`S_x, S_y, S_z`
    '''
    return S_x(s), S_y(s), S_z(s)

def SpinAngularMomenta(I, L, S):
    '''
    Returns angular momenta operators of a state given I, L, S in the tensor product basis.

    Parameters:
        I: nuclear spin quantum number of the atomic state
        L: orbital angular momentum quantum number
        S: spin quantum number of the state
    Returns:
        a tuple of angular momenta, :math:`((I_x, I_y, I_z), (L_x, L_y, L_z), (S_x, S_y, S_z))`
        (each a tuple of components of :math:`\\textbf{I}\\otimes\\textbf{L}\\otimes\\textbf{S}` as sparse matrices)
        in tensor product space.
    '''
    I_x = kron(kron(S_x(I), Id(2 * L + 1)), Id(2 * S + 1))
    I_y = kron(kron(S_y(I), Id(2 * L + 1)), Id(2 * S + 1))
    I_z = kron(kron(S_z(I), Id(2 * L + 1)), Id(2 * S + 1))
    L_x = kron(kron(Id(2 * I + 1), S_x(L)), Id(2 * S + 1))
    L_y = kron(kron(Id(2 * I + 1), S_y(L)), Id(2 * S + 1))
    L_z = kron(kron(Id(2 * I + 1), S_z(L)), Id(2 * S + 1))
    s_x = kron(kron(Id(2 * I + 1), Id(2 * L + 1)), S_x(S))
    s_y = kron(kron(Id(2 * I + 1), Id(2 * L + 1)), S_y(S))
    s_z = kron(kron(Id(2 * I + 1), Id(2 * L + 1)), S_z(S))
    IOperator = (I_x, I_y, I_z)
    LOperator = (L_x, L_y, L_z)
    SOperator = (s_x, s_y, s_z)
    return IOperator, LOperator, SOperator



import numpy as np
import warnings
import os
from sys import platform

if platform.startswith('win'):
    from ctypes import WinDLL, c_int, c_float, c_char_p, byref, POINTER, c_uint, \
        create_string_buffer, c_double, c_uint32


    class SifError:
        """
        This is a helper class to handle error codes produced by SIFReaderSDK.
        """
        ATSIF_SUCCESS = "ATSIF_SUCCESS"
        ATSIF_SIF_FORMAT_ERROR = "ATSIF_SIF_FORMAT_ERROR"
        ATSIF_NO_SIF_LOADED = "ATSIF_NO_SIF_LOADED"
        ATSIF_FILE_NOT_FOUND = "ATSIF_FILE_NOT_FOUND"
        ATSIF_FILE_ACCESS_ERROR = "ATSIF_FILE_ACCESS_ERROR"
        ATSIF_DATA_NOT_PRESENT = "ATSIF_DATA_NOT_PRESENT"
        ATSIF_P1INVALID = "ATSIF_P1INVALID"
        ATSIF_P2INVALID = "ATSIF_P2INVALID"
        ATSIF_P3INVALID = "ATSIF_P3INVALID"
        ATSIF_P4INVALID = "ATSIF_P4INVALID"
        ATSIF_P5INVALID = "ATSIF_P5INVALID"
        ATSIF_P6INVALID = "ATSIF_P6INVALID"
        ATSIF_P7INVALID = "ATSIF_P7INVALID"
        ATSIF_P8INVALID = "ATSIF_P8INVALID"

        ERROR_STR = {22002: ATSIF_SUCCESS,
                     22003: ATSIF_SIF_FORMAT_ERROR,
                     22004: ATSIF_NO_SIF_LOADED,
                     22005: ATSIF_FILE_NOT_FOUND,
                     22006: ATSIF_FILE_ACCESS_ERROR,
                     22007: ATSIF_DATA_NOT_PRESENT,
                     22101: ATSIF_P1INVALID,
                     22102: ATSIF_P2INVALID,
                     22103: ATSIF_P3INVALID,
                     22104: ATSIF_P4INVALID,
                     22105: ATSIF_P5INVALID,
                     22106: ATSIF_P6INVALID,
                     22107: ATSIF_P7INVALID,
                     22108: ATSIF_P8INVALID}

        @staticmethod
        def FromCode(errorCode):
            """Converts SIFReaderSDK error code to string.

            Parameters
            ----------
            errorCode : int
                Error code number

            Returns
            -------
            str
                Corresponding error string.

            """

            if errorCode in SifError.ERROR_STR:
                return SifError.ERROR_STR[errorCode]
            else:
                return "Unknown error (%d)"%(errorCode)

        @staticmethod
        def ProcessErrorCode(errorCode):
            """Processes error codes, raises RuntimeError if the error code indicate
            problems. In case of success, does nothing.

            Parameters
            ----------
            errorCode : int
                Error code number

            """
            errorStr = SifError.FromCode(errorCode)
            if errorStr != SifError.ATSIF_SUCCESS:
                raise RuntimeError("Error: %s"%(errorStr))


    # ------------------------------------------------------------------------------
    # AndorSifFile
    # ------------------------------------------------------------------------------

    class AndorSifFile(object):
        """
        This is main class to handle the reading of sif files. All the work is
        done in the constructor, the sif file is opened, data is read and the file
        is closed. This is adopted from https://github.com/ardiloot/AndorSifReader
        with modifications.

        Parameters:
            filename: str, Path to .sif file to read
        Attributes:
            signal : :class:`_SifFrame`
                Instance of the :class:`_SifFrame` helper class to store signal information.
            data : ndarray, Convenience property linking to signal.data.
            props : dict, Convenience property linking to signal.props.
        """

        # Read mode
        ATSIF_ReadAll = 0x40000000
        ATSIF_ReadHeaderOnly = 0x40000001

        # Data source
        ATSIF_Signal = 0x40000000
        ATSIF_Reference = 0x40000001
        ATSIF_Background = 0x40000002
        ATSIF_Live = 0x40000003
        ATSIF_Source = 0x40000004

        # Data types
        ATSIF_AT_8 = 0x40000000
        ATSIF_AT_U8 = 0x00000001
        ATSIF_AT_32 = 0x40000002
        ATSIF_AT_U32 = 0x40000003
        ATSIF_AT_64 = 0x40000004
        ATSIF_AT_U64 = 0x40000005
        ATSIF_Float = 0x40000006
        ATSIF_Double = 0x40000007
        ATSIF_String = 0x40000008

        # Converters
        ATSIF_CONV = {ATSIF_AT_8: np.int8,
                      ATSIF_AT_U8: np.uint8,
                      ATSIF_AT_32: np.int32,
                      ATSIF_AT_U32: np.uint32,
                      ATSIF_AT_64: np.int64,
                      ATSIF_AT_U64: np.uint64,
                      ATSIF_Float: np.float64,
                      ATSIF_Double: np.double,
                      ATSIF_String: lambda x: x.decode()
                      }

        # Calib axes
        ATSIF_CalibX = 0x40000000
        ATSIF_CalibY = 0x40000001
        ATSIF_CalibZ = 0x40000002

        # Property names
        ANTSIF_PROPS = [
            "Type",
            "Active",
            "Version",
            "Time",
            "FormattedTime",
            "FileName",
            "Temperature",
            "UnstabalizedTemperature",
            "Head",
            "HeadModel",
            "StoreType",
            "DataType",
            "SIDisplacement",
            "SINumberSubFrames",
            "PixelReadOutTime",
            "TrackHeight",
            "ReadPattern",
            "ReadPatternFullName",
            "ShutterDelay",
            "CentreRow",
            "RowOffset",
            "Operation",
            "Mode",
            "ModeFullName",
            "TriggerSource",
            "TriggerSourceFullName",
            "TriggerLevel",
            "ExposureTime",
            "Delay",
            "IntegrationCycleTime",
            "NumberIntegrations",
            "KineticCycleTime",
            "FlipX",
            "FlipY",
            "Clock",
            "AClock",
            "IOC",
            "Frequency",
            "NumberPulses",
            "FrameTransferAcquisitionMode",
            "BaselineClamp",
            "PreScan",
            "EMRealGain",
            "BaselineOffset",
            "SWVersion",
            "SWVersionEx",
            "MCP",
            "Gain",
            "VerticalClockAmp",
            "VerticalShiftSpeed",
            "OutputAmplifier",
            "PreAmplifierGain",
            "Serial",
            "DetectorFormatX",
            "DetectorFormatZ",
            "NumberImages",
            "NumberSubImages",
            "SubImageHBin",
            "SubImageVBin",
            "SubImageLeft",
            "SubImageRight",
            "SubImageTop",
            "SubImageBottom",
            "Baseline",
            "CCDLeft",
            "CCDRight",
            "CCDTop",
            "CCDBottom",
            "Sensitivity",
            "DetectionWavelength",
            "CountConvertMode",
            "IsCountConvert",
            "XAxisType",
            "XAxisUnit",
            "YAxisType",
            "YAxisUnit",
            "ZAxisType",
            "ZAxisUnit",
            "UserText",
            "IsPhotonCountingEnabled",
            "NumberThresholds",
            "Threshold1",
            "Threshold2",
            "Threshold3",
            "Threshold4",
            "AveragingFilterMode",
            "AveragingFactor",
            "FrameCount",
            "NoiseFilter",
            "Threshold",
            "TimeStamp"]

        def __init__(self, filename):
            c_int_p = POINTER(c_int)
            c_uint_p = POINTER(c_uint)
            c_float_p = POINTER(c_float)
            c_double_p = POINTER(c_double)

            self._filename = filename
            self.dll = WinDLL(os.path.join(os.path.dirname(__file__), 'ATSIFIO64.dll'))

            # Define library functions
            self.dll.ATSIF_SetFileAccessMode.argtypes = [c_int]
            self.dll.ATSIF_SetFileAccessMode.restype = c_uint

            self.dll.ATSIF_ReadFromFile.argtypes = [c_char_p]
            self.dll.ATSIF_ReadFromFile.restype = c_uint

            self.dll.ATSIF_CloseFile.argtypes = []
            self.dll.ATSIF_CloseFile.restype = c_uint

            self.dll.ATSIF_GetPropertyValue.argtypes = [c_int, c_char_p, c_char_p, c_uint]
            self.dll.ATSIF_GetPropertyValue.restype = c_uint

            self.dll.ATSIF_GetPropertyType.argtypes = [c_int, c_char_p, c_int_p]
            self.dll.ATSIF_GetPropertyType.restype = c_uint

            self.dll.ATSIF_GetFrameSize.argtypes = [c_int, c_uint_p]
            self.dll.ATSIF_GetFrameSize.restype = c_uint

            self.dll.ATSIF_GetFrame.argtypes = [c_int, c_uint, c_float_p, c_uint]
            self.dll.ATSIF_GetFrame.restype = c_uint

            # Open file
            self._Open()

            # Read contents
            try:
                self.signal = _SifFrames(self, self.ATSIF_Signal)
            finally:
                self._Close()

        def _Open(self):
            # Read all
            errorCode = self.dll.ATSIF_SetFileAccessMode(self.ATSIF_ReadAll)
            SifError.ProcessErrorCode(errorCode)

            # Open
            try:
                errorCode = self.dll.ATSIF_ReadFromFile(self._filename.encode())
            except OSError as e:
                if e.errno == 22:
                    raise IOError("Could not read %s"%(self._filename))
                raise

            SifError.ProcessErrorCode(errorCode)

        def _Close(self):
            errorCode = self.dll.ATSIF_CloseFile()
            SifError.ProcessErrorCode(errorCode)

        @property
        def data(self):
            return self.signal.data

        @property
        def props(self):
            return self.signal.props


    # ------------------------------------------------------------------------------
    # _SifFrames
    # ------------------------------------------------------------------------------

    class _SifFrames:
        """
        This is helper class to store single image/spectrum together with all the
        properties and calibration. The creation of this class is always through
        :class:`AndorSifFile`.

        Parameters:
            sif: :class:`AndorSifFile`, Reference to :class:`AndorSifFile`
            source: int, Andor code for signal/background/reference.
        Attributes:
            props: dict, Contain all the properties of the measurement
            data: ndarray, The data array to contain image/spectrum.
            n_frames: int, number of frames of the data (and other redundant attributes)

        """

        def __init__(self, sif, source):
            self._sif = sif
            self._source = source
            self.dll = self._sif.dll
            self.props = {}

            # Get number of pixels
            self._nrPixels = self._GetNumberOfPixels()

            # Read all properties
            for prop in AndorSifFile.ANTSIF_PROPS:
                self.props[prop] = self._GetProperty(prop)

            # Read images
            self.n_frames = self.props['NumberImages']
            self.width = (self.props['SubImageRight']-self.props['SubImageLeft']+1)//self.props['SubImageHBin']
            self.height = (self.props['SubImageTop']-self.props['SubImageBottom']+1)//self.props['SubImageVBin']
            self.data = np.zeros((self.n_frames, self.height, self.width))
            for i in range(self.n_frames):
                self.data[i] = np.array(self._ReadData(i)).reshape(self.height, self.width)

        def _GetNumberOfPixels(self):
            res = c_uint32()
            errorCode = self.dll.ATSIF_GetFrameSize(self._source, byref(res))
            SifError.ProcessErrorCode(errorCode)
            return res.value

        def _GetProperty(self, propertyName):
            resStr = create_string_buffer(260)
            errorCode = self.dll.ATSIF_GetPropertyValue(self._source, propertyName.encode(), resStr, len(resStr))
            SifError.ProcessErrorCode(errorCode)

            resType = c_int()
            errorCode = self.dll.ATSIF_GetPropertyType(self._source, propertyName.encode(), byref(resType))
            SifError.ProcessErrorCode(errorCode)

            res = self._ConvertToType(resStr.value, resType.value)
            return res

        def _ConvertToType(self, data, targetType):
            if targetType not in AndorSifFile.ATSIF_CONV:
                raise ValueError("Unknown target type.")
            res = AndorSifFile.ATSIF_CONV[targetType](data)
            return res

        def _ReadData(self, frameNum):
            data = (c_float*self._nrPixels)()
            errorCode = self.dll.ATSIF_GetFrame(self._source, frameNum, data, self._nrPixels)
            SifError.ProcessErrorCode(errorCode)
            return np.ctypeslib.as_array(data)

elif platform.startswith('lin'):
    warnings.warn('Andor *.sif files are read by sif-parser not by using ATSIFIO64.dll')
elif platform.startswith('dar'):
    warnings.warn('Andor *.sif files are read by sif-parser not by using ATSIFIO64.dll')
else:
    warnings.warn('Andor *.sif files could not be read in your OS as ATSIFIO64.dll is only available for windows.\
                   Try sif-parser package to read and implement it in __init__ of FluorescenceImage.')



from .fits import lorentzian, lorentzianFit, gaussian2DFit, gaussianFit, gaussian
import os
from .Images import approximatePositionOfTheCloud, rcParams
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import matplotlib.patches as patch
from scipy.constants import *
from scipy.special import wofz as w
from scipy.optimize import curve_fit
import numpy as np

def spectroscopy(ODimages, f, d=4, atom_loss=False, plot=True, fileNum='', lor_fit=False, savefig=False):
    '''
    Adds the OD of the pixels around the centre and uses the sum to plot the spectrum of the scan
    corresponding to the given frequencies. This is then fit to a gaussian (or a lorentzian)
    to find the center and width.

    Parameters:
        ODimages: ODimages extracted from ShadowImaging sequences
        f: array of frequencies for which the scan is done
        d: int, default is 4, size of the image area to consider around the centre of OD
        atom_loss: bool, default is False, flag to specify if atom loss spectroscopy is done
        plot: bool, default is True, flag to specify if the data has to be plotted
        fileNum: string, file number (plus any additional description) of the image file for which the analysis is done
        lor_fit: bool, default is False. This is a flag to fit the data to a lorentzian instead of a gaussian
        savefig: bool, default is False. Change it to true if you want to save the spectrum as .png

    Returns:
        a tuple, (amp, centre, sigma, offset) of the gaussian fit or (amp, centre, gamma, offset) of the lorentzian fit
    '''
    n = len(ODimages)
    f_smooth = np.linspace(f[0], f[-1]+(f[1]-f[0]), 200, endpoint=False)
    if n!=len(f):
        raise ValueError('No of images and no. of  frequencies are not equal')
    step = np.round(f[1]-f[0], 3)
    # finding approximate center of the cloud
    y, x, size = approximatePositionOfTheCloud(np.mean(ODimages, axis=0))
    index = []
    for i in range(n):
        index.append(np.sum(np.sum(ODimages[i][y+1-d:y+1+d, x+1-d:x+1+d])))
    maxODAt = np.argmax(index)
    minODAt = np.argmin(index)
    critical = maxODAt
    title = 'Max.'
    if lor_fit == False:
        try:
            if atom_loss==True:
                pOpt, pCov = gaussianFit(f, np.array(index), p0=[min(index), f[minODAt], 0.1, 0], plot=False, display=False)
                critical = minODAt
                title = 'Min.'
            else:
                pOpt, pCov = gaussianFit(f, np.array(index), p0=[max(index), f[maxODAt], 0.1, 0], plot=False, display=False)
        except RuntimeError:
            pOpt = []
    else:
        try:
            if atom_loss==True:
                pOpt, pCov = lorentzianFit(f, np.array(index), p0=[min(index), f[minODAt], 0.1, 0], plot=False, display=False)
                critical = minODAt
                title = 'Min.'
            else:
                pOpt, pCov = lorentzianFit(f, np.array(index), p0=[max(index), f[maxODAt], 0.1, 0], plot=False, display=False)
        except RuntimeError:
            pOpt = []
    if plot == True:
        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(9,3))
        i = ax[0].imshow(ODimages[critical])
        fig.colorbar(i, ax=ax[0])
        ax[0].scatter(x+1, y+1, marker='+', color='r')
        ax[0].set_title(title+' OD at: '+str(critical))
        rectangle = patch.Rectangle((x+1-d, y+1-d), 2*d, 2*d, linewidth=1,edgecolor='r',facecolor='none')
        ax[0].add_patch(rectangle)
        ax[0].grid(False)
        ax[1].plot(f, index, 'o')
        if pOpt!=[]:
            if lor_fit == True:
                ax[1].plot(f_smooth, lorentzian(f_smooth, *pOpt), 'k',
                           label='lor. fit:\n$\Gamma = {:.3f}$\n$f_0 = {:.3f}$'.format(abs(pOpt[2]), pOpt[1]))
            else:
                ax[1].plot(f_smooth, gaussian(f_smooth, *pOpt), 'k',
                           label='gaus. fit:\n $\sigma = {:.3f}$\n$f_0 = {:.3f}$'.format(abs(pOpt[2]), pOpt[1]))
            ax[1].legend()
        ax[1].set_ylabel('$\propto$ OD')
        ax[1].set_xlabel('$f$')
        ax[1].set_title(r'$f_{start}$ = '+str(f[0])+', $f_{step}$ = '+str(step)+', file = '+fileNum)
        plt.tight_layout()
        if savefig==True:
            try:
                plt.savefig('ColdTOFU_results/spectroscopy_result_for'+fileNum+'.png', transparent=True)
            except FileNotFoundError:
                pwd = os.getcwd()
                os.mkdir(os.path.join(pwd, 'ColdTOFU_results'))
                plt.savefig('ColdTOFU_results/spectroscopy_result_for'+fileNum+'.png', transparent=True)
    return pOpt, index # amp, centre, gamma, offset

def bv(f, f0, b0, T, s, offset):
    '''
    Function representing convolution of the lorentzian line shape of the red transition and gaussian maxwell
    distribution. This is taken from 5.13 from Chang chi's thesis and added the effect of saturation parameter.

    Args:
        f: numpy.array, frequency vector
        f0: float, resonance frequency or centre of the spectrum
        b0: float, optical depth at resonance at zero temperature
        T: float, temperature in micro K
        s: float, saturation parameter of the probe, :math:`I/I_s`.
        offset: float, offset

    Returns:
        optical depth for frequencies f in the shape f.
    '''
    gamma = 2*pi*7.5*milli
    vavg = np.sqrt(T*micro*Boltzmann/(87*m_p))
    k = 2*pi/(689*nano)
    x = w((2*pi*(f-f0) + 1j*gamma*np.sqrt(1+s)/2)*1e6/(np.sqrt(2)*k*vavg))
    return b0*np.sqrt(pi/8)*(1/(1+s))*(gamma*np.sqrt(1+s)*1e6/(k*vavg))*x.real+offset

def bvFit(f, array, p0=None, bounds=None):
    '''
    Function to fit spectroscopy data to real line shape of the transition.

    Args:
        f: numpy.array, frequency vector
        array: float, optical depth at scan frequencies f
        p0: initial guess for the fit as [f_0, b_0, T (in :math:`\mu K`), s]
        bounds: bounds for the fit as ([lower bounds], [upper bounds])

    Returns:
        a tuple with optimized parameters and covariance ex: (pOpt, pCov)
    '''
    pOpt, pCov = curve_fit(bv, f, array, p0, bounds=bounds)
    return pOpt, pCov

def spectroscopyFaddeva(ODimages, f, plot=True, atom_loss=False, fileNum='', savefig=False):
    '''
    Fits od images to a gaussian and uses its amplitude to plot the spectrum of the scan
    corresponding to the given frequencies. This is fit to :math:`b_v(\delta)` from Chang Chi's thesis to extract
    temperature in addition to center.

    Parameters:
        ODimages: ODimages extracted from ShadowImaging sequences
        f: array of frequencies for which the scan is done
        atom_loss: bool, default is False, flag to specify if atom loss spectroscopy is done
        plot: bool, default is True to specify if the data has to be plotted
        fileNum: string, file number (plus any additional description) of the image file for which the analysis is done.
        savefig: bool, default is False. Change it to true if you want to save the spectrum as .png

    Returns:
        a tuple, (centre, :math:`b_0`, T, s) of fit to :math:`b_v(\delta)`
    '''
    n = len(ODimages)
    imaging_params = rcParams().params
    f_smooth = np.linspace(f[0], f[-1] + (f[1] - f[0]), 100, endpoint=False)
    if n!=len(f):
        raise ValueError('No of images and no. of  frequencies are not equal')
    step = np.round(f[1]-f[0], 3)
    y, x, size = approximatePositionOfTheCloud(np.mean(ODimages, axis=0))
    index = []
    title = 'Max.'
    p0 = [1, x, y, size, size, 0, 0.1]
    bounds = ([-3, x-4, y-4, 0.2*size, 0.2*size, 0, -0.1], [4.5, x+4, y+4, 1.8*size, 1.8*size, 6.28, 0.3])
    for i in range(n):
        amp, xo, yo, sx, sy, theta, offset = gaussian2DFit(ODimages[i], p0, bounds, plot=False)[0]
        index.append(amp)
    if atom_loss==False:
        maxODAt = np.argmax(index)
        criticalOD = np.max(index)
    else:
        maxODAt = np.argmin(index)
        criticalOD = np.min(index)
        title = 'Min.'
    try:
        s = imaging_params['saturation'] # this should be for atom-loss probe in case of atom-loss spectroscopy not imaging probe
        b = ([min(f), -300, 0.01,  0.9*s, -100], [max(f), 300, 20,  1.1*s, 100])
        pOpt, pCov = bvFit(f, np.array(index), p0=[f[maxODAt], criticalOD, 5, s, 0], bounds=b)
    except RuntimeError:
        pOpt = [0, 0, 0, 0]
    T = pOpt[2]
    pixelSize = imaging_params['pixelSize']
    magnification = imaging_params['magnification']
    binning = imaging_params['binning']
    sizeFactor = pixelSize*binning/magnification
    

    if plot == True:
        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(9,3))
        if size != 0:
            i = ax[0].imshow(ODimages[maxODAt][y-3*int(size):y+3*int(size), x-3*int(size):x+3*int(size)])
        else:
            i = ax[0].imshow(ODimages[maxODAt])
        scalebar = AnchoredSizeBar(ax[0].transData, 2, str(np.round(2*sizeFactor/1e-6, 1))+r'$\mu$m',
                                   'lower right', color='white', frameon=False,size_vertical=0.2)

        ax[0].add_artist(scalebar)
        ax[0].set_title(title+' OD at: '+str(maxODAt))
        fig.colorbar(i, ax=ax[0])     
        ax[0].grid(False)
        ax[1].plot(f, index, 'o')
        if pOpt!=[]:
            ax[1].plot(f_smooth, bv(f_smooth, *pOpt), 'k', label=r'T='+str(np.round(T, 1))+'$\mu$K \n'+
                                                  '$b_0(0)$='+str(np.round(pOpt[1], 2))+'\n'+
                                                  '$f_0$ = '+str(np.round(pOpt[0], 3))+'\n'+
                                                  's = '+str(np.round(pOpt[3], 2)))
            ax[1].legend(loc='upper right')
        ax[1].set_ylabel('OD') # ignore comment \\times \sigma_x \\times \sigma_y
        ax[1].set_xlabel('$f$')
        ax[1].set_title(r'$f_{start}$ = '+str(f[0])+', $f_{step}$ = '+str(step)+', file = '+fileNum)
        plt.tight_layout()
        if savefig==True:
            try:
                plt.savefig('ColdTOFU_results/spectroscopy_Faddeva_result_for'+fileNum+'.png', transparent=True)
            except FileNotFoundError:
                pwd = os.getcwd()
                os.mkdir(os.path.join(pwd, 'ColdTOFU_results'))
                plt.savefig('ColdTOFU_results/spectroscopy_Faddeva_result_for'+fileNum+'.png', transparent=True)
    return pOpt, index



import os
from scipy.io import loadmat
from numpy.fft import fft, fftfreq
import numpy as np

def picoMatRead(filePath, channels=['A']):
    """
    Reads .mat files exported using picoscope software.

    Parameters:
        filePath: string, path of the file to read.
        channels: list of strings, channels of the file to read. Ex: `['A', 'B']`

    Returns:
        a touple, (time array, list of data arrays)
    """
    file = loadmat(filePath)
    tStep = file['Tinterval']
    N = file['Length'][0, 0]
    t = np.arange(0,N)*tStep
    result = np.zeros((len(channels), N))
    for i, ch in enumerate(channels):
        result[i] = file[ch][:,0]
    return t, result


def PSD(path, avg=30, channels=['A']):
    """
    Reads multiple waveforms .mat files folder exported using picoscope software and
     alculates power spectral density (PSD).

    Parameters:
        path: string, path of the folder to read.
        avg: int, number of waveforms in the folder to average.
        channels: list of strings, channels of the file to read. Ex: `['A', 'B']`

    Returns:
        a touple, (frequencies array, list of PSD arrays in the order of channels)
    """
    files = os.listdir(path)
    first = loadmat(os.path.join(path, files[0]))
    tStep = first['Tinterval']
    N = first['Length'][0,0]
    frequencies = fftfreq(N, tStep)[0,:int(N/2)]
    result = np.zeros((len(channels), int(N/2)))
    for f in files[:avg]:
        data = loadmat(os.path.join(path, f))
        for i, ch in enumerate(channels):
            result[i] += 2*abs(fft(data[ch][:,0])[:N//2]/N)**2
    result /= avg
    result[0] = result[0]/2
    return frequencies, result


def RIN(path, avg=30, channel='A'):
    """
    Reads multiple waveforms .mat files folder exported using picoscope software and
    calculates relative intensity noise (RIN).

    Parameters:
        path: string, path of the folder to read.
        avg: int, number of waveforms in the folder to average.
        channels: list of strings, channels of the file to read. Ex: `['A', 'B']`

    Returns:
        a touple, (frequencies array, list of RIN arrays in the order of channels)
    """
    files = os.listdir(path)
    first = loadmat(os.path.join(path, files[0]))
    tStep = first['Tinterval']
    N = first['Length'][0,0]
    frequencies = fftfreq(N, tStep)[0,:int(N/2)]
    inLoop = []
    for f in files[:avg]:
        data = loadmat(os.path.join(path, f))
        inLoop.append(2*abs(fft(data[channel][:,0])[:N//2]/N)**2/np.mean(data[channel][:, 0])**2)
    inLoopRIN = np.mean(np.array(inLoop), axis=0)
    inLoopRIN[0], inLoopRIN[-1] = inLoopRIN[0]/2, inLoopRIN[-1]/2
    return frequencies, 10*np.log10(inLoopRIN)




import numpy as np
from scipy.sparse import csr_matrix



def a(N=5):
    '''
    Matrix representation of bosonic destruction operator.

    Parameters:
        n: destroy boson at n in an N state Fock space
        N: dimensions of the Fock space, default 5
    Returns:
        a sparse matrix representation of a of shape (N, N) in Fock basis, {:math:`|n_1, n_2, n_3, ..., N\\rangle`}.
    '''
    if N<1:
        raise ValueError('number of bosons must be greater than 1!')
    else:
        n_n = np.arange(0, N, 1)
        rows = np.arange(0, N-1, 1)
        cols = np.arange(1, N, 1)
        values = np.array([np.sqrt(n) for n in n_n[1:]])
        spar = csr_matrix((values, (rows, cols)), shape=(int(N), int(N)))
        return spar

def a_dag(N=5):
    '''
    Matrix representation of bosonic creation operator.

    Parameters:
        n: create a boson at n in an N state Fock space
        N: dimensions of the Fock space, default 5
    Returns:
        a sparse matrix representation of a of shape (N, N) in Fock basis, {:math:`|n_1, n_2, n_3, ..., N\\rangle`}.
    '''
    if N<1:
        raise ValueError('number of bosons must be greater than 0 and less than 1!')
    else:
        n_n = np.arange(0, N, 1)
        rows = np.arange(1, N, 1)
        cols = np.arange(0, N-1, 1)
        values = np.array([np.sqrt(n+1) for n in n_n[:-1]])
        spar = csr_matrix((values, (rows, cols)), shape=(int(N), int(N)))
        return spar


import numpy as np
from math import *
from ..picoMatTools import picoMatRead

class TSeries(object):
    def __init__(self, dataPath, channel, f0, Amplitude=None, offset=None):
        data = picoMatRead(dataPath, channels=[channel])
        self.time = data[0][0] # time axis
        self.f0 = f0 # nominal frequency of the ref. osc.
        self.sampleTime = self.time[1]-self.time[0]
        if Amplitude==None:
            if offset==None:
                beat = data[1][0] - np.mean(data[1][0])
                max_sig = np.max(abs(beat))
                phase = np.arccos(beat/max_sig)
                self.angle = np.unwrap(phase) # phase fluctuation
        else:
            beat = data[1][0] - offset
            phase = np.arccos(beat/Amplitude)
            self.angle = np.unwrap(phase)
    def fracFreq(self):
        deltaFreq = np.gradient(self.angle, self.time)/(2*pi) # freq. fluctuation
        return deltaFreq/self.f0
    def timeDiff(self):
        self.x = self.angle/(2*pi*self.f0) # time difference
        return self.x
    def allenVar(self, m):
        x = self.timeDiff()
        tau = self.time[m]-self.time[0] #sampling time in sec
        y = np.zeros(int(len(x)/m))
        for i in range(len(x)//m-1):
            y[i] = (x[m*i+m]-x[m*i])/tau
        s = 0
        for i in range(len(y)-1):
            s += (y[i+1]-y[i])**2
        return s/(2*(len(y)-1))
    def allenDev(self, m):
        return np.sqrt(self.allenVar(m))
    def oAllenVar(self, m):
        x = self.timeDiff()
        tau = self.time[m]-self.time[0] #sampling time in sec
        y = np.zeros(len(x)-m)
        for i in range(len(x)-m-1):
            y[i] = (x[i+m]-x[i])/tau
        s = 0
        for i in range(len(y)-1):
            s += (y[i+1]-y[i])**2
        return s/(2*(len(y)-1))
    def oAllenDev(self, m):
        return np.sqrt(self.oAllenVar(m))


from .Images import *
from .AndorSifReader import *
from .fits import *
from .jupyterTools import *
from .numberOfAtoms import *
from .picoMatTools import *
from .sigma import *
from .spectrum import *
from .Physics import *


import numpy as np
import scipy.linalg as ln


class PauliBasis():
    '''
    A class to represent Pauli's spin basis.
    '''
    def _init__(self):
        return self

    def matrices(self):
        '''

        Returns:
            a list of 4 :math:`2\\times 2` matrices corresponding to :math:`I_2, \\sigma_x, \\sigma_y, \\sigma_z`
        '''
        sigx = np.array([[0, 1], [1, 0]])
        sigy = np.array([[0, -1j], [1j, 0]])
        sigz = np.array([[1, 0], [0, -1]])
        Id = np.eye(2)
        return np.array([Id, sigx, sigy, sigz])

    def decompose(self, A):
        '''
        Decomposes given non-singular :math:`2\\times 2` matrix in the Pauli basis.

        Args:
            A: :math:`2\\times 2` matrix to be decomposed.

        Returns:
            list of 4 components.
        '''
        assert np.array(A).shape==(2,2), "Not a 2x2 matrix to decompose in pauli basis."
        l = self.matrices()
        M = np.zeros((4,4), dtype=complex)
        for i in range(4):
            for j in range(4):
                M[i, j] = l[j].flatten()[i]
        try:
            return ln.inv(M)@(np.array(A).flatten())
        except ln.LinAlgError:
            raise(ln.LinAlgError, "Matrix singular! Unique decomposition doesn't exist!")

class GellMannBasis():
    def _init__(self):
        self.structureConstants()
        return self

    def matrices(self):
        I = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        l1 = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
        l2 = np.array([[0, -1j, 0], [1j, 0, 0], [0, 0, 0]])
        l3 = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 0]])
        l4 = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])
        l5 = np.array([[0, 0,-1j], [0, 0, 0], [1j, 0, 0]])
        l6 = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]])
        l7 = np.array([[0, 0, 0], [0, 0, -1j], [0, 1j, 0]])
        l8 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -2]])/np.sqrt(3)
        return np.array([I, l1, l2, l3, l4, l5, l6, l7, l8])

    def structureConstants(self):
        f = np.zeros((8, 8, 8), dtype=complex)
        l = self.matrices()
        for i in range(8):
            for j in range(8):
                for k in range(8):
                    f[i,j,k] = -1j*np.trace(np.dot(np.dot(l[i+1], l[j+1]), l[k+1]) - np.dot(np.dot(l[j+1], l[i+1]), l[k+1]))/4
        self.f = f
        return self.f

    def structureConstant(self, i, j, k):
        return self.f[i-1, j-1, k-1]

    def decompose(self, A):
        l = self.matrices()
        result = np.zeros(9, dtype=complex)
        result[0] = np.trace(A)/3
        A = A-np.trace(A)*np.eye(3)/3
        for i in range(1, len(l)):
            result[i] = np.trace(A@l[i])/2 # Since, Gellmann matrices are trace normalized to 2
        return result

    def compose(self, decomposition):
        assert len(decomposition)==9, 'Provided decomposition is not of size 9!'
        l = self.matrices()
        result = np.zeros((3,3), dtype=complex)
        for i in range(9):
            result += decomposition[i]*l[i]
        return result